/**************************************************************************//**
*   @file   ad5791.c
*   @brief  Implementation of AD5791 Driver for Microblaze processor.
*   @author ATofan (alexandru.tofan@analog.com)
*
*******************************************************************************
* Copyright 2011(c) Analog Devices, Inc.
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*  - Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in
*    the documentation and/or other materials provided with the
*    distribution.
*  - Neither the name of Analog Devices, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived
*    from this software without specific prior written permission.
*  - The use of this software may or may not infringe the patent rights
*    of one or more patent holders.  This license does not release you
*    from the requirement that you obtain separate licenses from these
*    patent holders to use this software.
*  - Use of the software either in source or binary form, must be run
*    on or directly connected to an Analog Devices Inc. component.
*
* THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY
* AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*******************************************************************************
*   SVN Revision: 468
******************************************************************************/
#define AD5791_GLOBALS

/*****************************************************************************/
/* Include Files                                                             */
/*****************************************************************************/
#include "ad5791.h"
#include "../drivers/spi.h"
#include "../drivers/gpio.h"

/*****************************************************************************/
/* Private variables                                                         */
/*****************************************************************************/
unsigned char   txBuffer[4]     = {0,};     // Tx SPI Buffer.
unsigned char   rxBuffer[4]     = {0,};     // Rx SPI Buffer.
static char     firstInit       = 1;        // Used to control the first initialization of some variables

/*****************************************************************************/
/* Aliases for uC/Probe data                                                 */
/*****************************************************************************/
volatile int*   ad5791Error             = &AD5791_Data[0].iData;    // Indicates a SPI connection error
volatile int*   ad5791DacValue          = &AD5791_Data[1].iData;    // Value to be written in DAC Register
volatile int*   ad5791ClrCodeValue      = &AD5791_Data[2].iData;    // Value to be written in Clear Code Register
volatile int*   ad5791ClrCodeValueGet   = &AD5791_Data[3].iData;    // Value read from Clear Code Register
volatile int*   ad5791DacValueGet       = &AD5791_Data[4].iData;    // Value read from DAC Register
volatile int*   ad5791CodeC2            = &AD5791_Data[5].iData;    // DAC Register Coding in Twos Complement
volatile int*   ad5791CodeBO            = &AD5791_Data[6].iData;    // DAC Register Coding in Binary Offset
volatile float* ad5791VOut              = &AD5791_Data[7].fData;    // AD5791 Output Voltage
volatile char*  ad5791LedSdodis         = &AD5791_Data[8].chData;   // SDODIS bit in Control Register
volatile char*  ad5791LedBin2sc         = &AD5791_Data[9].chData;   // BIN2SC bit in Control Register
volatile char*  ad5791LedDactri         = &AD5791_Data[10].chData;  // DACTRI bit in Control Register
volatile char*  ad5791LedOpgnd          = &AD5791_Data[11].chData;  // OPGND bit in Control Register
volatile char*  ad5791LedRbuf           = &AD5791_Data[12].chData;  // RBUF bit in Control Register
volatile char*  ad5791LedReset          = &AD5791_Data[13].chData;  // RESET bit in Software Control Register
volatile char*  ad5791LedClr            = &AD5791_Data[14].chData;  // CLR bit in Software Control Register
volatile char*  ad5791LedLdac           = &AD5791_Data[15].chData;  // LDAC bit in Software Control Register
volatile char*  ad5791LoadBtn           = &AD5791_Data[16].chData;  // Load DAC button
volatile char*  ad5791RbufBtn           = &AD5791_Data[17].chData;  // Set/Clear RBUF button
volatile float* ad5791VrefP             = &AD5791_Data[18].fData;   // Positive voltage reference value
volatile float* ad5791VrefN             = &AD5791_Data[19].fData;   // Negative voltage reference value
volatile int*   ad5791SoftCtrlReg       = &AD5791_Data[20].iData;   // Software Control Value

/**************************************************************************//**
 * @brief Initialize SPI and Initial Values for AD5791 Board.
 *
 * @param None.
 *
 * @return success if the communication with the AD5791 works correct;
 *         fail otherwise.
******************************************************************************/
int AD5791_Init( void )
{
    if(firstInit)
    {
        *ad5791CodeBO = 0;
        *ad5791CodeC2 = 1;
        firstInit = !firstInit;
    }
    int spiData = 0;
    int cmdCheck;
    int retValue = FALSE;
    // setup SPI interface
    SPI_Init(AD5791_SPI, 0, 0, 1);                                  // Setup the SPI interface.
    // setup AD5791 control register
    spiData = ( AD5791_CMD_WR_CTRL << AD5791_CMD_POSITION ); // Write to the control register
    if ( *ad5791CodeC2 == 1)
    {
        spiData += ( CTRL_LIN_COMP );   // Activate Linear Compensation
        if( *ad5791RbufBtn == 1 )
        {
            spiData += CTRL_RBUF;       // If RBUF button is active, Set RBUF bit
        }
        AD5791_Set( &spiData );         // Send the data
        spiData = AD5791_Get();         // Read the control register
        cmdCheck = ( CTRL_LIN_COMP );   // Create the Check Sequence
        if( *ad5791RbufBtn == 1 )
        {
            cmdCheck += CTRL_RBUF;
        }
        if (spiData & cmdCheck)
        {
            // check the echo
            retValue = TRUE;
        }
    }
    if ( *ad5791CodeBO == 1)
    {
        spiData += ( CTRL_LIN_COMP | CTRL_BIN2SC ); // Activate Linear Compensation and Binary Offset Coding
        if( *ad5791RbufBtn == 1 )
        {
            spiData += CTRL_RBUF;                   // If RBUF button is active, Set RBUF bit
        }
        AD5791_Set( &spiData );                     // Send the data
        spiData = AD5791_Get();                     // Read the control register
        cmdCheck = ( CTRL_LIN_COMP | CTRL_BIN2SC ); // Create the Check Sequence
        if( *ad5791RbufBtn == 1 )
        {
            cmdCheck += CTRL_RBUF;
        }
        if (spiData & cmdCheck)
        {
            // check the echo
            retValue = TRUE;
        }
    }
    *ad5791Error = !retValue;
    AD5791_SetLed();

    return( retValue );
}

/**************************************************************************//**
 * @brief Set uC/Probe GUI specific LED's
 *
 * @param None.
 *
 * @return None.
******************************************************************************/
void AD5791_SetLed( void )
{
    int spiData = 0;
    int ctrlReg = 0;
    // Read the Control Register
    spiData = ( AD5791_CMD_RD_CTRL << AD5791_CMD_POSITION );
    AD5791_Set( &spiData );
    ctrlReg = AD5791_Get();
    // Set GUI LED's according to the register values
    *ad5791LedSdodis = ctrlReg & CTRL_SDODIS;
    *ad5791LedBin2sc = ctrlReg & CTRL_BIN2SC;
    *ad5791LedDactri = ctrlReg & CTRL_DACTRI;
    *ad5791LedOpgnd = ctrlReg & CTRL_OPGND;
    *ad5791LedRbuf = ctrlReg & CTRL_RBUF;
    *ad5791LedReset = *ad5791SoftCtrlReg & SOFT_CTRL_RESET;
    *ad5791LedClr = *ad5791SoftCtrlReg & SOFT_CTRL_CLR;
    *ad5791LedLdac = *ad5791SoftCtrlReg & SOFT_CTRL_LDAC;
}

/**************************************************************************//**
 * @brief Write to the DAC register using two's complement
 *
 * @param dacValue
 *
 * @return None.
******************************************************************************/
void AD5791_SetDacC2( int dacValue )
{
    int spiData = 0;
    // Load data in DAC Register
    spiData = ( AD5791_CMD_WR_DAC << AD5791_CMD_POSITION );
    spiData += ( dacValue << AD5791_DAC_DATA_POSITION );
    AD5791_Set( &spiData );
}

/**************************************************************************//**
 * @brief Write to the DAC register using binary offset
 *
 * @param dacValue
 *
 * @return None.
******************************************************************************/
void AD5791_SetDacBO( int dacValue )
{
    int spiData = 0;
    // Load data in DAC Register
    spiData = ( AD5791_CMD_WR_DAC << AD5791_CMD_POSITION );
    spiData += ( dacValue << AD5791_DAC_DATA_POSITION );
    AD5791_Set( &spiData );
}

/**************************************************************************//**
 * @brief Write to the Clear Code register
 *
 * @param clrCodeValue
 *
 * @return None.
******************************************************************************/
void AD5791_SetClrCode( int clrCodeValue )
{
    int spiData = 0;
    // Load data in Clear Code Register
    spiData = ( AD5791_CND_WR_CLR_CODE << AD5791_CMD_POSITION );
    spiData += ( clrCodeValue << AD5791_CLR_CODE_DATA_POSITION );
    AD5791_Set( &spiData );
}

/**************************************************************************//**
 * @brief Write to the Software Control register
 *
 * @param softCtrlValue (data bit number in register to be set)
 *
 * @return None.
******************************************************************************/
void AD5791_SetSoftCtrl( int softCtrlValue )
{
    int spiData = 0;
    // Load data in Software Control Register
    spiData = ( AD5791_CMD_WR_SOFT_CTRL << AD5791_CMD_POSITION );
    switch( softCtrlValue )
    {
        case 0:
            spiData += ( SOFT_CTRL_LDAC );
            break;
        case 1:
            spiData += ( SOFT_CTRL_CLR );
            break;
        case 2:
            spiData += ( SOFT_CTRL_RESET );
            break;
        default:
            break;
    }
    // Write data in the Software Control Register
    AD5791_Set( &spiData );
    // Save Software Control Register Value (it cannot be read via SPI)
    *ad5791SoftCtrlReg = spiData;
}

/**************************************************************************//**
 * @brief Read from the DAC register
 *
 * @param None.
 *
 * @return dacValue.
******************************************************************************/
int AD5791_GetDac( void )
{
    int spiData = 0;
    int dacValue = 0;
    // Read from the DAC Register
    spiData = ( AD5791_CMD_RD_DAC << AD5791_CMD_POSITION );
    AD5791_Set( &spiData );
    dacValue = AD5791_Get();
    // Select data bits only
    dacValue = ( dacValue & 0xFFFFC );
    dacValue = ( dacValue >> 2 );

    return( dacValue );
}

/**************************************************************************//**
 * @brief Read from the Control register
 *
 * @param None.
 *
 * @return ctrlValue.
******************************************************************************/
int AD5791_GetCtrl( void )
{
    int spiData = 0;
    int ctrlValue = 0;
    // Read from the Control Register
    spiData = ( AD5791_CMD_RD_CTRL << AD5791_CMD_POSITION );
    AD5791_Set( &spiData );
    ctrlValue = AD5791_Get();

    return( ctrlValue );
}

/**************************************************************************//**
 * @brief Read from the Clear Code register
 *
 * @param None.
 *
 * @return clrCodeValue.
******************************************************************************/
int AD5791_GetClrCode( void )
{
    int spiData = 0;
    int clrCodeValue = 0;
    // Read from the Clear Code Register
    spiData = ( AD5791_CMD_RD_CLR_CODE << AD5791_CMD_POSITION );
    AD5791_Set( &spiData );
    clrCodeValue = AD5791_Get();

    return( clrCodeValue );
}

/**************************************************************************//**
 * @brief Read and interpret data
 *
 * @param None.
 *
 * @return vOut.
******************************************************************************/
float AD5791_SetGet( void )
{
    float division_constant;
    float vOut = 0;
    int adcValue;
    int iData = ( AD5791_GetCtrl() & ~(CTRL_RBUF) );
    // Check status of RBUF button and act accordingly
    if( *ad5791RbufBtn )
    {
        iData += CTRL_RBUF;
    }
    AD5791_Set( &iData );
    // Check status of Load Button and act accordingly
    if( *ad5791LoadBtn == 1)
    {
        AD5791_SetSoftCtrl( 0 );                    // Set LDAC bit
    }
    else
    {
        AD5791_SetSoftCtrl( 4 );                    // Clear LDAC bit
    }
    // Set and check Clear Code register Value
    AD5791_SetClrCode( *ad5791ClrCodeValue );
    *ad5791ClrCodeValueGet = AD5791_GetClrCode();
    // Check DAC Register Coding option and act accordingly
    if ( *ad5791CodeC2 == 1)
    {
        AD5791_SetDacC2( *ad5791DacValue );         // Write to the DAC register using two's complement coding
    }
    if ( *ad5791CodeBO == 1)
    {
        AD5791_SetDacBO( *ad5791DacValue );         // Write to the DAC register using binary offset coding
    }
    // Read DAC Register Value
    *ad5791DacValueGet = AD5791_GetDac();
    // Calculate Vout for the following parameters
    // ad5791VrefP = User selected
    // ad5791VrefN = User selected
    // division constant = (ad5791VrefP - ad5791VrefN)/2^20
    division_constant = ( *ad5791VrefP - *ad5791VrefN ) / 1048575;
    if ( *ad5791CodeBO)
    {
        vOut = division_constant * *ad5791DacValue + *ad5791VrefN;
    }
    else
    {
        if (*ad5791DacValue &  0x80000)
        {
            adcValue = ~(*ad5791DacValue) + 1;
            adcValue = adcValue & 0xFFFFF;
            adcValue = adcValue + 524288 ;
            vOut = -(division_constant * adcValue + *ad5791VrefN);

        }
        else
        {
            adcValue = *ad5791DacValue + 524288;
            vOut = division_constant * adcValue + *ad5791VrefN;
        }

    }

    return( vOut );
}

/**************************************************************************//**
 * @brief Send SPI Data
 *
 * @param value.
 *
 * @return TRUE.
******************************************************************************/
int AD5791_Set( int *value )
{
    txBuffer[0] = ( unsigned char ) ( ( *value & 0xff0000 ) >> 16 ); // Arrange the
    txBuffer[1] = ( unsigned char ) ( ( *value & 0x00ff00 ) >> 8 );  // data to be sent
    txBuffer[2] = ( unsigned char ) ( *value & 0x0000ff );           // in 8 bit packets
    if (!SPI_TransferData(AD5791_SPI, 3, (char *) txBuffer, 0, NULL, 1))
    {
        SPI_Init(AD5791_SPI, 0, 0, 1);                                  // Setup the SPI interface.
    }
    return(TRUE);
}

/**************************************************************************//**
 * @brief Read SPI Data
 *
 * @param None.
 *
 * @return Retrieved data.
******************************************************************************/
int AD5791_Get( void )
{
    int returnVal = 0;
    txBuffer[0] = ( unsigned char ) 0;
    txBuffer[1] = ( unsigned char ) 0;
    txBuffer[2] = ( unsigned char ) 0;
    if (!SPI_TransferData(AD5791_SPI, 0, NULL , 3, (char *) rxBuffer, 1))
    {
        SPI_Init(AD5791_SPI, 0, 0, 1);                                  // Setup the SPI interface.
    }
    returnVal = (int) ( rxBuffer[0] << 16 );
    returnVal  += (int) ( rxBuffer[1] << 8 );
    returnVal += rxBuffer[2];

    return( returnVal );
}

/**************************************************************************//**
 * @brief Update the device's monitored and controlled parameters
 *
 * @param None
 *
 * @return Always returns success
******************************************************************************/
int AD5791_UpdateParameters(void)
{
if ( *ad5791CodeBO == *ad5791CodeC2 )
    {
        *ad5791CodeC2 = !*ad5791CodeBO;
        if (TRUE != AD5791_Init())
        {
            return(FALSE);
        }
    }
    *ad5791VOut = AD5791_SetGet();
    AD5791_SetLed();

    return(TRUE);
}
